#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "resource.h"

#define MAX_PROFILE_PATH 4096
static char profile_path[MAX_PROFILE_PATH] = "";

void multiply_matrix_by_vector(double* left, double* right, double* result)
{
	// matrix * vertical vector
	result[0] = left[0*3+0]*right[0] + left[0*3+1]*right[1] + left[0*3+2]*right[2];
	result[1] = left[1*3+0]*right[0] + left[1*3+1]*right[1] + left[1*3+2]*right[2];
	result[2] = left[2*3+0]*right[0] + left[2*3+1]*right[1] + left[2*3+2]*right[2];
}

void multiply_vector_by_matrix(double* left, double* right, double* result)
{
	// horizontal vector * matrix
	result[0] = left[0]*right[0*3+0] + left[1]*right[1*3+0] + left[2]*right[2*3+0];
	result[1] = left[0]*right[0*3+1] + left[1]*right[1*3+1] + left[2]*right[2*3+1];
	result[2] = left[0]*right[0*3+2] + left[1]*right[1*3+2] + left[2]*right[2*3+2];
}

void multiply_matrix_by_matrix(double* left, double* right, double* result)
{
	result[0*3+0] = left[0*3+0]*right[0*3+0] + left[0*3+1]*right[1*3+0] + left[0*3+2]*right[2*3+0];
	result[0*3+1] = left[0*3+0]*right[0*3+1] + left[0*3+1]*right[1*3+1] + left[0*3+2]*right[2*3+1];
	result[0*3+2] = left[0*3+0]*right[0*3+2] + left[0*3+1]*right[1*3+2] + left[0*3+2]*right[2*3+2];
	result[1*3+0] = left[1*3+0]*right[0*3+0] + left[1*3+1]*right[1*3+0] + left[1*3+2]*right[2*3+0];
	result[1*3+1] = left[1*3+0]*right[0*3+1] + left[1*3+1]*right[1*3+1] + left[1*3+2]*right[2*3+1];
	result[1*3+2] = left[1*3+0]*right[0*3+2] + left[1*3+1]*right[1*3+2] + left[1*3+2]*right[2*3+2];
	result[2*3+0] = left[2*3+0]*right[0*3+0] + left[2*3+1]*right[1*3+0] + left[2*3+2]*right[2*3+0];
	result[2*3+1] = left[2*3+0]*right[0*3+1] + left[2*3+1]*right[1*3+1] + left[2*3+2]*right[2*3+1];
	result[2*3+2] = left[2*3+0]*right[0*3+2] + left[2*3+1]*right[1*3+2] + left[2*3+2]*right[2*3+2];
}

void inverse_matrix(double* a, double* b)
{
	double det;
	det =   a[0*3+0]*a[1*3+1]*a[2*3+2] - a[0*3+0]*a[2*3+1]*a[1*3+2] + a[1*3+0]*a[2*3+1]*a[0*3+2]
	      - a[1*3+0]*a[0*3+1]*a[2*3+2] + a[2*3+0]*a[0*3+1]*a[1*3+2] - a[2*3+0]*a[1*3+1]*a[0*3+2];
	b[0*3+0] =  (a[1*3+1]*a[2*3+2] - a[2*3+1]*a[1*3+2]) / det;
	b[0*3+1] = -(a[0*3+1]*a[2*3+2] - a[2*3+1]*a[0*3+2]) / det;
	b[0*3+2] =  (a[0*3+1]*a[1*3+2] - a[1*3+1]*a[0*3+2]) / det;
	b[1*3+0] = -(a[1*3+0]*a[2*3+2] - a[2*3+0]*a[1*3+2]) / det;
	b[1*3+1] =  (a[0*3+0]*a[2*3+2] - a[2*3+0]*a[0*3+2]) / det;
	b[1*3+2] = -(a[0*3+0]*a[1*3+2] - a[1*3+0]*a[0*3+2]) / det;
	b[2*3+0] =  (a[1*3+0]*a[2*3+1] - a[2*3+0]*a[1*3+1]) / det;
	b[2*3+1] = -(a[0*3+0]*a[2*3+1] - a[2*3+0]*a[0*3+1]) / det;
	b[2*3+2] =  (a[0*3+0]*a[1*3+1] - a[1*3+0]*a[0*3+1]) / det;
}

void CalcValue(HWND hDlg)
{
	char buf[256];
	double chroma_matrix[9], tmp_matrix[9], cnv_matrix[9], inv_matrix[9];
	double ref_vector[3], tgt_vector[3], tmp_vector[3], rgb_vector[3];
	double gamma, x, y;
	
	// reference white vector setup
	GetDlgItemText(hDlg, IDC_REF_X, buf, 255);
	ref_vector[0] = atof(buf);
	GetDlgItemText(hDlg, IDC_REF_Y, buf, 255);
	ref_vector[1] = atof(buf);
	ref_vector[2] = 1.0 - ref_vector[0] - ref_vector[1];
	
	// phosphor matrix setup
	GetDlgItemText(hDlg, IDC_RED_X, buf, 255);
	x = atof(buf);
	GetDlgItemText(hDlg, IDC_RED_Y, buf, 255);
	y = atof(buf);
	chroma_matrix[0*3+0] = x/y;
	chroma_matrix[0*3+1] = 1;
	chroma_matrix[0*3+2] = (1-x-y)/y;
	
	GetDlgItemText(hDlg, IDC_GREEN_X, buf, 255);
	x = atof(buf);
	GetDlgItemText(hDlg, IDC_GREEN_Y, buf, 255);
	y = atof(buf);
	chroma_matrix[1*3+0] = x/y;
	chroma_matrix[1*3+1] = 1;
	chroma_matrix[1*3+2] = (1-x-y)/y;
	
	GetDlgItemText(hDlg, IDC_BLUE_X, buf, 255);
	x = atof(buf);
	GetDlgItemText(hDlg, IDC_BLUE_Y, buf, 255);
	y = atof(buf);
	chroma_matrix[2*3+0] = x/y;
	chroma_matrix[2*3+1] = 1;
	chroma_matrix[2*3+2] = (1-x-y)/y;
	
	// tone reproduction parameter setup
	GetDlgItemText(hDlg, IDC_GAMMA, buf, 255);
	gamma = atof(buf);
	
	// target white vector setup
	GetDlgItemText(hDlg, IDC_TARGET_X, buf, 255);
	tgt_vector[0] = atof(buf);
	GetDlgItemText(hDlg, IDC_TARGET_Y, buf, 255);
	tgt_vector[1] = atof(buf);
	tgt_vector[2] = 1.0 - tgt_vector[0] - tgt_vector[1];
	
	inverse_matrix(chroma_matrix, tmp_matrix);
	multiply_vector_by_matrix(ref_vector, tmp_matrix, tmp_vector);
	cnv_matrix[0*3+0] = chroma_matrix[0*3+0] * tmp_vector[0] / ref_vector[1];
	cnv_matrix[1*3+0] = chroma_matrix[1*3+0] * tmp_vector[1] / ref_vector[1];
	cnv_matrix[2*3+0] = chroma_matrix[2*3+0] * tmp_vector[2] / ref_vector[1];
	cnv_matrix[0*3+1] = chroma_matrix[0*3+1] * tmp_vector[0] / ref_vector[1];
	cnv_matrix[1*3+1] = chroma_matrix[1*3+1] * tmp_vector[1] / ref_vector[1];
	cnv_matrix[2*3+1] = chroma_matrix[2*3+1] * tmp_vector[2] / ref_vector[1];
	cnv_matrix[0*3+2] = chroma_matrix[0*3+2] * tmp_vector[0] / ref_vector[1];
	cnv_matrix[1*3+2] = chroma_matrix[1*3+2] * tmp_vector[1] / ref_vector[1];
	cnv_matrix[2*3+2] = chroma_matrix[2*3+2] * tmp_vector[2] / ref_vector[1];
	inverse_matrix(cnv_matrix, inv_matrix);
	
	multiply_vector_by_matrix(tgt_vector, inv_matrix, rgb_vector);
	double max_rgb = (rgb_vector[0] > rgb_vector[1]) ? rgb_vector[0] : rgb_vector[1];
	max_rgb = (rgb_vector[2] > max_rgb) ? rgb_vector[2] : max_rgb;
	rgb_vector[0] = rgb_vector[0] / max_rgb;
	rgb_vector[1] = rgb_vector[1] / max_rgb;
	rgb_vector[2] = rgb_vector[2] / max_rgb;
	
	sprintf_s(buf, 256, "%7.5f", pow(rgb_vector[0], 1/gamma));
	SetDlgItemText(hDlg, IDC_BALANCE_R, buf);
	sprintf_s(buf, 256, "%7.5f", pow(rgb_vector[1], 1/gamma));
	SetDlgItemText(hDlg, IDC_BALANCE_G, buf);
	sprintf_s(buf, 256, "%7.5f", pow(rgb_vector[2], 1/gamma));
	SetDlgItemText(hDlg, IDC_BALANCE_B, buf);
}

void ResetValue(HWND hDlg)
{
	// reference white vector setup
	SetDlgItemText(hDlg, IDC_REF_X, "0.3127");
	SetDlgItemText(hDlg, IDC_REF_Y, "0.3290");
	
	// phosphor matrix setup
	SetDlgItemText(hDlg, IDC_RED_X, "0.640");
	SetDlgItemText(hDlg, IDC_RED_Y, "0.330");
	
	SetDlgItemText(hDlg, IDC_GREEN_X, "0.300");
	SetDlgItemText(hDlg, IDC_GREEN_Y, "0.600");
	
	SetDlgItemText(hDlg, IDC_BLUE_X, "0.150");
	SetDlgItemText(hDlg, IDC_BLUE_Y, "0.060");
	
	// tone reproduction parameter setup
	SetDlgItemText(hDlg, IDC_GAMMA, "2.2");
	
	// target white vector setup
	SetDlgItemText(hDlg, IDC_TARGET_X, "0.3127");
	SetDlgItemText(hDlg, IDC_TARGET_Y, "0.3290");
	
	CalcValue(hDlg);
}

void GetProfilePath(HWND hDlg)
{
	OPENFILENAME ofn;
	
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hDlg;
	ofn.lpstrFile = profile_path;
	ofn.lpstrFilter = TEXT("ICC profiles (*.icc;*.icm)\0*.icc;*.icm\0all files (*.*)\0*.*\0\0");
	ofn.nMaxFile = MAX_PROFILE_PATH;
	ofn.Flags = OFN_FILEMUSTEXIST;
	GetOpenFileName(&ofn);
}

#define U32BE(x) ((*(unsigned char*)(x)<<24)|(*((unsigned char*)(x)+1)<<16)|(*((unsigned char*)(x)+2)<<8)|*((unsigned char*)(x)+3))
#define U16BE(x) ((*(unsigned char*)(x)<<8)|*((unsigned char*)(x)+1))
#define S16BE(x) ((*(signed char*)(x)<<8)|*((unsigned char*)(x)+1))

void LoadProfile(HWND hDlg)
{
	FILE* fp;
	unsigned int ntag, tag, index, fmt, i;
	double X, Y, Z, Num, Gamma;
	unsigned char* buf;
	long profile_size;
	char buf2[256];
	
	if( (fp = fopen(profile_path, "rb")) == NULL ) return;
	fseek(fp, 0, SEEK_END);
	profile_size = ftell(fp);
	if( (buf = (unsigned char*)malloc(profile_size)) == NULL ) return;
	rewind(fp);
	fread(buf, 1, profile_size, fp);
	
	if( U32BE(buf + 0x0c) != 0x6D6E7472U || // 'mntr'
		U32BE(buf + 0x10) != 0x52474220U || // 'RGB '
		U32BE(buf + 0x14) != 0x58595A20U || // 'XYZ '
		U32BE(buf + 0x24) != 0x61637370U )  // 'acsp'
		return;
	ntag = U32BE(buf+128);
	
	for( i = 0; i < ntag; i++ ){
		tag   = U32BE(buf + 132 + i * 12);
		index = U32BE(buf + 136 + i * 12);
		fmt   = U32BE(buf + index);
		switch( fmt ){
		case 0x58595A20U: // 'XYZ '
			X = S16BE(buf + index + 8 ) + (U16BE(buf + index + 10) / 65536.0);
			Y = S16BE(buf + index + 12) + (U16BE(buf + index + 14) / 65536.0);
			Z = S16BE(buf + index + 16) + (U16BE(buf + index + 18) / 65536.0);
			switch( tag ){
			case 0x77747074U: // 'wtpt'
				sprintf_s(buf2, 256, "%7.5f", X / (X+Y+Z));
				SetDlgItemText(hDlg, IDC_REF_X, buf2);
				sprintf_s(buf2, 256, "%7.5f", Y / (X+Y+Z));
				SetDlgItemText(hDlg, IDC_REF_Y, buf2);
				break;
			case 0x7258595AU: // 'rXYZ'
				sprintf_s(buf2, 256, "%7.5f", X / (X+Y+Z));
				SetDlgItemText(hDlg, IDC_RED_X, buf2);
				sprintf_s(buf2, 256, "%7.5f", Y / (X+Y+Z));
				SetDlgItemText(hDlg, IDC_RED_Y, buf2);
				break;
			case 0x6758595AU: // 'gXYZ'
				sprintf_s(buf2, 256, "%7.5f", X / (X+Y+Z));
				SetDlgItemText(hDlg, IDC_GREEN_X, buf2);
				sprintf_s(buf2, 256, "%7.5f", Y / (X+Y+Z));
				SetDlgItemText(hDlg, IDC_GREEN_Y, buf2);
				break;
			case 0x6258595AU: // 'bXYZ'
				sprintf_s(buf2, 256, "%7.5f", X / (X+Y+Z));
				SetDlgItemText(hDlg, IDC_BLUE_X, buf2);
				sprintf_s(buf2, 256, "%7.5f", Y / (X+Y+Z));
				SetDlgItemText(hDlg, IDC_BLUE_Y, buf2);
				break;
			default:
				break;
			}
			break;
		case 0x63757276U: // 'curv'
			Num = U32BE(buf + index + 8);
			Gamma = U16BE(buf + index + 12) / 256.0;
			switch(tag){
			case 0x72545243U: // 'rTRC'
				break;
			case 0x67545243U: // 'gTRC'
				if(Num == 1){
					sprintf_s(buf2, 256, "%7.5f", Gamma);
					SetDlgItemText(hDlg, IDC_GAMMA, buf2);
				}
				break;
			case 0x62545243U: // 'bTRC'
				break;
			default:
				break;
			}
			break;
		case 0x70617261U: // 'para'
			break;
		default:
			break;
		}
	}
	CalcValue(hDlg);
}

void CenteringDialog(HWND hDlg)
{
	RECT r;
	GetWindowRect(hDlg, &r);
	int wscr = GetSystemMetrics(SM_CXSCREEN);
	int hscr = GetSystemMetrics(SM_CYSCREEN);
	int w = r.right - r.left;
	int h = r.bottom - r.top;
	MoveWindow(hDlg, (wscr - w)>>1, (hscr - h)>>1, w, h, FALSE);
}

LRESULT CALLBACK DlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM)
{
	int id = LOWORD(wParam);
	
	switch(message){
	case WM_CLOSE:
		EndDialog(hDlg, IDOK);
		return TRUE;
		
	case WM_INITDIALOG:
		CenteringDialog(hDlg);
		ResetValue(hDlg);
		return TRUE;
		
	case WM_COMMAND:
		switch (id){
		case IDC_LOAD:
			GetProfilePath(hDlg);
		case IDC_RELOAD:
			LoadProfile(hDlg);
			return TRUE;
		case IDC_CALC:
			CalcValue(hDlg);
			return TRUE;
		case IDC_CLOSE:
			EndDialog(hDlg, IDOK);
			return TRUE;
		default:
			break;
		}
		break;
	}
	return FALSE;
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE, LPTSTR, int)
{
	DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)DlgProc);
	return 0;
}
